/*
 *  plex86: run multiple x86 operating systems concurrently
 *  Copyright (C) 2000 Kevin P. Lawton
 *
 *  fault.c:  fault/int handlers for VM monitor
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */


#include "plex86.h"
#define IN_MONITOR_SPACE
#include "monitor.h"


/* The monitor stack frame.  When an exception or interrupt occurrs
 * during the execution of either guest or monitor code, the following
 * values are pushed.
 * 
 * [gs]
 * [fs]       These are pushed by the CPU _only if_ transitioning from
 * [ds]       guest code running in v86 mode.
 * [es]
 *
 * ss
 * esp
 * eflags    Values pushed by the CPU and interrupt stub.  To simplify
 * cs        things, the stub pushes an error of zero for those
 * eip       events which don't naturally cause an error push, and
 * error     also pushes the vector of the exception/interrupt.
 * vector
 *
 * eax
 * ecx
 * edx       General registers, pushed with a PUSHA instruction,
 * ebx       by code below.
 * <esp>
 * ebp
 * esi
 * edi
 *
 * es
 * ds        Segment selectors, pushed by code below.
 * fs
 * gs
 */

/* Info on hack for transitioning to 16-bit guest stack segments.  For
 * the cases of returning to a guest being monitored in either v86-mode,
 * or PM with a 32-bit SS, the full 32 bits of ESP are honored from the
 * monitor stack image.  But when performing an IRET back to guest code
 * monitored in PM * with a 16-bit stack segment, only the lower 16 bits
 * of ESP are initialized.  The upper 16bits remain from the previous and
 * unrelated monitor ESP values.  To handle restoring the full 32 bits of
 * the guest ESP as expected, a temporary monitor stack segment is created,
 * with adjusted SS.base and ESP values such that the same linear offsets
 * are generated for the parameters (guest register values) on the
 * monitor stack, but with the upper 16 bits of ESP being identical
 * to those expected by the guest.  This temporary segment is used only
 * for the purposes of the IRET.  During the next event generated while
 * the guest is executing, the normal SS:ESP values are reloaded from
 * the monitor TSS.
 *
 * The only issue is if an exception is generated while trying to
 * IRET to the guest code.  In this case, we have to fix the SS:ESP
 * first, as these values will not have been reloaded as they usually
 * are in a transition from guest (ring3) to monitor (ring0).
 */



static inline Bit32u read_cr2( void )
{
    Bit32u cr2;
    asm volatile("movl %%cr2, %0" : "=r" (cr2));
    return cr2;
}


asm (
".text                  \n\t"

/* __handle_fault:  This is called by all of the monitor's fault handler
 *     stubs.  A fault could have originated from execution of the guest
 *     (due to virtualization conditions or natural fault generation) or
 *     from the monitor (currently only due to bugs in the monitor).
 */
".globl __handle_fault  \n\t"
"__handle_fault:        \n\t"
"  pushal               \n\t" /* Save general registers */
"  pushl %es            \n\t" /* Save segment registers */
"  pushl %ds            \n\t"
"  pushl %fs            \n\t"
"  pushl %gs            \n\t"

"  movl  64(%esp), %eax     \n\t" /* EFLAGS pushed by CPU from fault */
"  andl  $0x20000, %eax     \n\t" /* Check EFLAGS.VM bit */
"  jnz   __fault_from_guest \n\t" /* V86 mode means fault was from guest */
"  movl  60(%esp), %eax     \n\t" /* CS pushed by CPU from fault */
"  andl  $3, %eax           \n\t" /* Check CS.RPL bits */
"  jz    __fault_from_mon   \n\t" /* RPL0 means from monitor */

/* We have determined that the fault was from guest code.  Prepare
 * to call the monitor C code to do most of the fault handling.
 */
"__fault_from_guest:    \n\t"
"  movl  %ss, %eax      \n" /* Copy SS into DS/ES */
"  movl  %eax, %ds      \n"
"  movl  %eax, %es      \n"
"  cld                  \n" /* gcc-compiled code needs this */
"  call handle_fault    \n" /* Call the C monitor fault handler */
".globl __ret_to_guest  \n" /* Fault handled, work back to guest */
"__ret_to_guest:        \n"
"  call sbe             \n" /* Prepare for return to guest */
"  cmpl $0x1, %eax      \n" /* What mode is guest monitored in? */
"  jb   __ret_to_v86    \n" /* case 0: guest monitored in v86 mode */
"  jnz  __ret_to_pmss16 \n" /* case 2: guest monitored in PM w/ small SS */
                            /* case 1: guest monitored in PM w/ big SS */

/* Return to the guest, monitored in PM with a 32-bit SS.  Simply
 * restore registers from the monitor stack.  The full 32-bit guest ESP
 * value is honored.
 */
"__ret_to_pmss32:       \n"
"  popl  %gs            \n" /* Restore guest segments */
"  popl  %fs            \n"
"  popl  %ds            \n"
"  popl  %es            \n"
"  popal                \n" /* Restore guest general registers */
"  addl  $8, %esp       \n" /* Ignore vector and error dwords */
"  iret                 \n" /* Resume execution of guest */


/* __fault_from_mon:  Currently this means a monitor bug.  It's often
 *   due to an attempt to IRET to the guest with bad selector/descriptor
 *   values setup by the monitor.  Because of the SS16 hack, if an IRET
 *   back to guest code never completes and generates an exception, our
 *   SS:ESP values are still modified and not reloaded by the exception
 *   mechanism since no privilege level change occurred.  So before
 *   turning control over to the C code, we must check for this condition
 *   and restore SS:ESP.  Then we can reload the data segments from SS.
 */

"__fault_from_mon:               \n\t"
"  movl  %esp, %ebx              \n\t" /* Get nexus page address */
"  andl  $0xfffff000, %ebx       \n\t"
"  movl  $__SSNormal, %edx       \n\t" /* Create a pointer to SSNormal */
"  subl  $__nexus_start, %edx    \n\t"
"  ss; movl  (%ebx,%edx), %ecx   \n\t" /* Get SSNormal */
"  movl  %ss, %eax               \n\t" /* Get current SS */
"  cmpw  %cx, %ax                \n\t" /* Compare SS with SSNormal */
"  je    __mon_SS_ESP_restored   \n\t" /* If same, then no restore needed */

"  movl  $__espUpperNormal, %edx \n\t" /* Get pointer to espUpperNormal */
"  subl  $__nexus_start, %edx    \n\t"
"  ss; movl  (%ebx,%edx), %eax   \n\t" /* Get espUpperNormal */
"  andl  $0x0000ffff, %esp       \n\t" /* Clear upper ESP bits */
"  orl   %eax, %esp              \n\t" /* Restore upper ESP bits */
"  movl  %ecx, %ss               \n\t" /* Restore SS */

"__mon_SS_ESP_restored:          \n\t"
"  movl  %ss, %eax               \n\t" /* Copy monitor SS to DS/ES */
"  movl  %eax, %ds               \n\t"
"  movl  %eax, %es               \n\t"
"  cld                           \n\t" /* gcc-compiled code needs this */
"  call handle_mon_fault         \n\t" /* Call C code for real work */
"  jmp  __ret_to_monitor         \n\t" /* Event handled, back to monitor */
                                       /* (Currently never gets here) */



/* Return to guest, monitored in PM with a 16-bit SS.  Only the lower
 * 16 bits of ESP will be honored.  The upper 16 bits of ESP are actually
 * passed through (left) from the unrelated monitor ESP value, when
 * transitioning from a 32-bit SS (which is the case with the monitor)
 * to a 16-bit SS.  So what we do is create an alternate monitor stack
 * segment such that the upper 16 bits are equivalent to those expected
 * by the guest, by adjusting the SS.base value.  This alternate segment
 * is really only used for the purposes of an IRET back to the guest.
 */
"__ret_to_pmss16:                     \n\t"
"  movl  %esp, %ebx                   \n\t" /* Get nexus page addr */
"  andl  $0xfffff000, %ebx            \n\t"
"  movl  $__espUpper16BitSSHack, %edx \n\t" /* Get ptr to nexus field */
"  subl  $__nexus_start, %edx         \n\t"
"  movl  (%ebx,%edx), %eax            \n\t" /* Access nexus field */
"  movl  $__SS16BitSSHack, %edx       \n\t" /* Get ptr to nexus field */
"  subl  $__nexus_start, %edx         \n\t"
"  movl  (%ebx,%edx), %ecx            \n\t" /* Access nexus field */
"  andl  $0x0000ffff, %esp            \n\t" /* Clear upper ESP bits */
"  orl   %eax, %esp                   \n\t" /* Use alternate ESP upper bits */
"  movl  %ecx, %ss                    \n\t" /* Use alternate SS segment */
"  popl  %gs            \n\t" /* Restore guest segments */
"  popl  %fs            \n\t"
"  popl  %ds            \n\t"
"  popl  %es            \n\t"
"  popal                \n\t" /* Restore guest general registers */
"  addl  $8, %esp       \n\t" /* Ignore vector and error dwords */
"  iret                 \n\t" /* Resume execution of guest */

/* Return to the guest, monitored in v86 mode.  Simply restore registers
 * from the monitor stack.  The full 32-bit guest ESP value is honored
 * when returning to v86 mode code.
 */
"__ret_to_v86:          \n"
"  addl $16, %esp       \n" /* Ignore, segs are in IRET area for v86 */
"  popal                \n" /* Restore guest general registers */
"  addl $8, %esp        \n" /* Ignore vector and error dwords */
"  iret                 \n" /* Resume execution of guest */

/* Return to monitor.  An event occurred while monitor code was executing.
 * Simply restore state from the monitor stack.
 */
"__ret_to_monitor:      \n"
"  popl  %gs            \n" /* Restore monitor segments */
"  popl  %fs            \n"
"  popl  %ds            \n"
"  popl  %es            \n"
"  popal                \n" /* Restore monitor general registers */
"  addl  $8, %esp       \n" /* ignore vector and error dwords */
"  iret                 \n" /* Resume execution of monitor */


/*
 * Hardware interrupt handler stub
 */
".globl __handle_int    \n" /* Return to monitor code */
"__handle_int:          \n"
"  pushal               \n" /* Save guest general registers */
"  pushl %es            \n" /* Save guest segment registers */
"  pushl %ds            \n"
"  pushl %fs            \n"
"  pushl %gs            \n"

"  movl  %ss, %eax      \n" /* Copy SS into DS/ES */
"  movl  %eax, %ds      \n"
"  movl  %eax, %es      \n"
"  cld                  \n" /* gcc-compiled code needs this */
"  call handle_int      \n" /* monitor interrupt handler */
"  cmpl $0x1, %eax      \n" /* Was interrupt generated from monitor code? */
"  je   __ret_to_monitor\n" /* Yes, so return to monitor code */
"  jmp  __ret_to_guest  \n" /* No, so return to guest code */
);



  unsigned
sbe( guest_context_t context, v86_sregs_t v86_sregs )
{
    nexus_t *nexus = (nexus_t *) (((Bit32u) &context) & 0xfffff000);
    vm_t    *vm    = (vm_t *) nexus->vm;

    Bit32u   guest_laddr, guest_paddr, guest_ppage_index;
    unsigned seg32;
    unsigned icache_index;
    unsigned pdi, pti;
    Bit8u    temp8;
    pageEntry_t pte_new;
    Bit32u private_codepage_laddr, gerror;
    unsigned us, rw;
    page_t *monPTbl;
    unsigned pt_index;
    unsigned m, ret;

#if ANAL_CHECKS
    if ( !context.eflags.fields.vm && !(context.cs & 0x0003) )
      monpanic(vm, "sbe: monitor code!\n");
#endif


#if ANAL_CHECKS
    if ( !context.eflags.fields.if_ )
       monpanic(vm, "sbe: guest IF=0\n");

    if ( context.eflags.raw & (
         FLG_IOPL | /* FLG_NT | FLG_RF | FLG_VM | FLG_AC | */
         FLG_VIF | FLG_VIP) )
       monpanic(vm, "sbe: eflags=0x%x\n", context.eflags);
#endif


    STI();

loop:

    guest_cycles_elapsed(vm);

    /* Past execution (emulation) could have created a need for
     * a remap of the guest in the monitor space
     */
    if (vm->modeChange)
      monModeChange(vm);

    /* Check for asynchronous events, inhibits, traps etc before
     * executing guest code
     */
    emulate_async_checks(vm);

    /* +++ Look again here???  Remove check above??? */
    if (vm->modeChange)
      monModeChange(vm);

/* Notes:
 *   if (vm->guest_cpu.inhibit_mask || vm->guest_cpu.debug_trap)
 *     then have to make sure inhibits are honored.
 */

    /* See if cosimulation demands the guest takes an interrupt
     * at this point.  This would be the case if it is being
     * driven as a slave simulator.
     */
    if (vm->dbg_force_int) {
      unsigned vector = vm->dbg_force_int & 0xff;
      if (vm->guest_cpu.inhibit_mask || vm->guest_cpu.debug_trap)
        monpanic(vm, "sbe: force_int w/ IM || DT\n");
      vm->dbg_force_int = 0;
      emulate_interrupt(vm, vector);
      sysReqComplete(vm);
      /* An interrupt may put the VM in a state which requires
       * remapping.  Repeat the process
       */
      goto loop;
      }

    switch (vm->executeMethod) {
      case RunGuestNMethodExecute:
        /* Really, should copy guest TF to monitor TF here.  For
         * now just turn off TF always.  Need to check what case
         * set it in handler when enabled.
         */
        if (vm->guest_cpu.inhibit_mask || vm->guest_cpu.debug_trap)
          monpanic(vm, "sbe: MethodExecute: IM || DT\n");
        context.eflags.fields.tf = 0;
        break;

      case RunGuestNMethodEmulate:
        emulate_instr(vm, &context, 4);
        /* Above call will decrement executeN if execution completes */
        goto loop;

      case RunGuestNMethodBreakpoint:
        /* Regardless of guest TF, we need to set monitor TF to generate
         * a breakpoint after the next instruction executes
         */
        context.eflags.fields.tf = 1;
        break;

      default:
        monpanic(vm, "sbe: Method default\n");
      }

    cache_sreg(vm, SRegCS);
    if (context.eip > vm->guest_cpu.desc_cache[SRegCS].limit_scaled)
      monpanic(vm, "sbe: guest EIP > CS.limit\n");
    guest_laddr = vm->guest_cpu.desc_cache[SRegCS].base + context.eip;
    us = (G_GetCPL(vm)==3);
    rw = 0; /* need at least read priv for code */
    switch ( m = map_guest_laddr(vm, guest_laddr, &guest_ppage_index,
                             us, rw, PageUsageVCode, &gerror) ) {
      case MapLinOK:
      case MapLinAlreadyMapped:
        break;
      case MapLinMonConflict:
        monpanic(vm, "sbe: MapLinMonConflict:\n");
      case MapLinPPageOOB:
        monpanic(vm, "sbe: MapLinPPageOOB:\n");
      case MapLinException:
        /* Executing code at this location would generate a #PF */
        vm->guest_cpu.cr2 = guest_laddr;
        emulate_exception(vm, ExceptionPF, gerror);
        /* An exception means a possible monitor remap request, mode
         * change etc.  We need to repeat this process.
         */
        goto loop;
      case MapLinEmulate:
        monpanic(vm, "sbe: MapLinEmulate:\n");
        emulate_instr(vm, &context, 5);
        goto loop;
      default:
        monpanic(vm, "sbe: MapLin: default case:\n");
      }

    if ( vm->vOpcodeMap ) {
      seg32 = vm->guest_cpu.desc_cache[SRegCS].desc.d_b;
      guest_paddr = (guest_ppage_index<<12) | (guest_laddr & 0xfff);
  
      /* The monitor linear address of the real code page */
      vm->codepage_laddr = Guest2Monitor(vm, guest_laddr & 0xfffff000);
  
      /* The address of the PTE pointing to the real code page */
      pdi = guest_laddr >> 22;
      pti = (guest_laddr >> 12) & 0x3ff;
  
      pt_index = getMonPTi(vm, pdi, 15);
      monPTbl = &vm->guest.addr.page_tbl[pt_index];
  
      vm->codepage_pte_p = &monPTbl->pte[pti];
      vm->codepage_pte_saved = *vm->codepage_pte_p;
  
      icache_index = prescan(vm, (Bit8u *) vm->codepage_laddr, guest_paddr,
                             seg32, G_GetCPL(vm), 1);
      if (icache_index == (unsigned)-1) {
        /* If the very first instruction is virtualized, there is no
         * sense trying to execute it.  It will only result in emulation
         * of the instruction, so just do that now.
         */
        emulate_instr(vm, &context, 5);
        goto loop;
        }
  
      /* Update timestamp for latest time of guest execution */
      vm->code_cache[icache_index].ts.guest_executed = vm_rdtsc();
  
  
      /* Split I&D TLB trick.  Our scan-before-execute technique */
      /* creates a private modified (virtualized) copy of the current */
      /* code page.  We want to be able to execute from this private */
      /* code page, yet allow memory accesses to see the original */
      /* code/data page so code does not detect our changes.  Since */
      /* the IA32 processor does not offer a native mechanism to */
      /* differentiate between read and code accesses within a page, */
      /* we make use of the split I&D TLB nature.  What we need to */
      /* do is to get the TLB to load a value with user privilege */
      /* into the I TLB cache, then reset the corresponding page */
      /* table entry to contain supervisor privilege.  Thus code */
      /* will continue to execute based on the user privilege level, */
      /* yet data accesses will fetch the entry from the actual */
      /* page tables and cause an exception. */
  
      /* The linear address of the virtualized code page */
      private_codepage_laddr =
        ((Bit32u)vm->guest.addr.icache) + (icache_index<<12);
  
      /* The temporary PTE pointing to the virtualized page */
      pte_new.fields.base = vm->pages.icache[icache_index];
      pte_new.fields.avail = 0;
      pte_new.fields.G = 0;      /* not global           */
      pte_new.fields.PS = 0;     /* (unused in pte)      */
      pte_new.fields.D = 0;      /* clean                */
      pte_new.fields.A = 0;      /* not accessed         */
      pte_new.fields.PCD = 0;    /* normal caching       */
      pte_new.fields.PWT = 0;    /* normal write-back    */
      pte_new.fields.US = 1;     /* user                 */
      pte_new.fields.RW = 0;     /* read-only            */
      pte_new.fields.P = 1;      /* present in memory    */
      }

    if (GetMonMode(vm) == MonModeVM) {
      /* If guest code is monitored in V86M, then it is important
       * that we load the context area with the correct selectors,
       * even though they are currently virtualized.
       */

      if ( vm->selectorInEmu & (1 << SRegES) )
        context.es = vm->guest_cpu.selector[SRegES].raw;
      if ( vm->selectorInEmu & (1 << SRegCS) )
        context.cs = vm->guest_cpu.selector[SRegCS].raw;
      if ( vm->selectorInEmu & (1 << SRegSS) )
        context.ss = vm->guest_cpu.selector[SRegSS].raw;
      if ( vm->selectorInEmu & (1 << SRegDS) )
        context.ds = vm->guest_cpu.selector[SRegDS].raw;
      if ( vm->selectorInEmu & (1 << SRegFS) )
        context.fs = vm->guest_cpu.selector[SRegFS].raw;
      if ( vm->selectorInEmu & (1 << SRegGS) )
        context.gs = vm->guest_cpu.selector[SRegGS].raw;

      /* Since we are executing guest code using v86 mode, we will
       * allow natural selector reads, and segment reloads to occur
       * unvirtualized.  Thus, we need to mark the fact that selectors
       * and descriptor cache values are no longer consistent with
       * those stored in the guest_cpu area.  
       */
      vm->selectorInEmu   = 0;
      vm->descriptorInEmu = 0;
      vm->segmentUpdated  = 0;

/* +++ we could eliminate double copy of selectors here and below */
      }
    else if (GetGuestMode(vm) == GuestModeRM) {
if ( (vm->selectorInEmu & 0x3f) != 0x3f )
  monpanic(vm, "sbe: MonModeVM, sInEmu=0x%x\n",
    vm->selectorInEmu);

      /* MonModePMR3, GuestModeRM.  Since the monitor is not running
       * guest code in VM, there must be non-compatible values in
       * the descriptor caches.  In this case we are virtualizing
       * segment registers accesses, and using PM selectors.  Update
       * virtualized descriptors based on updated guest values.  In
       * this case, there is no notion of a null selector.
       */
      if ( vm->segmentUpdated ) {
        unsigned i;
        for (i=0; i<6; i++) {
          if (vm->guest_cpu.desc_cache[i].valid) {
            vm->guest.addr.gdt[4+i] = vm->guest_cpu.desc_cache[i].desc;
            vm->guest.addr.gdt[4+i].dpl = 3;
            }
          else {
            vm->guest.addr.gdt[4+i] = nullDescriptor;
            }
          }
        vm->segmentUpdated = 0;
        }
      }
    else { /* GuestModePMR3 */

if ( (vm->selectorInEmu & 0x3f) != 0x3f )
  monpanic(vm, "sbe: MonModeVM, sInEmu=0x%x\n",
    vm->selectorInEmu);

if (!vm->vOpcodeMap && vm->segmentUpdated)
  monpanic(vm, "sbe0: sbe==0, segmentUpdated=0x%x\n", vm->segmentUpdated);

      if ( vm->segmentUpdated ) {
        unsigned i;

/* +++ similar code in mon-mode.c */
        /* If the descriptor caches for each segment register is
         * invalid, then set the virtualized selector to NULL so
         * an exception is not generated upon return to the guest.
         * Otherwise, point them at virtualized descriptors representing
         * the shadow cache values of each segment register.
         */
        if ( vm->segmentUpdated & (1<<SRegES) ) {
          if ( !vm->guest_cpu.desc_cache[SRegES].valid )
            vm->guest.addr.guest_context->es = nullSelector.raw;
          else
            vm->guest.addr.guest_context->es = Selector(4, 0, RPL3);
          }
     
        if ( vm->segmentUpdated & (1<<SRegCS) ) {
          if ( !vm->guest_cpu.desc_cache[SRegCS].valid )
            vm->guest.addr.guest_context->cs = nullSelector.raw;
          else
            vm->guest.addr.guest_context->cs = Selector(5, 0, RPL3);
          }

        if ( vm->segmentUpdated & (1<<SRegSS) ) {
          if ( !vm->guest_cpu.desc_cache[SRegSS].valid )
            vm->guest.addr.guest_context->ss = nullSelector.raw;
          else
            vm->guest.addr.guest_context->ss = Selector(6, 0, RPL3);
          }

        if ( vm->segmentUpdated & (1<<SRegDS) ) {
          if ( !vm->guest_cpu.desc_cache[SRegDS].valid )
            vm->guest.addr.guest_context->ds = nullSelector.raw;
          else
            vm->guest.addr.guest_context->ds = Selector(7, 0, RPL3);
          }

        if ( vm->segmentUpdated & (1<<SRegFS) ) {
          if ( !vm->guest_cpu.desc_cache[SRegFS].valid )
            vm->guest.addr.guest_context->fs = nullSelector.raw;
          else
            vm->guest.addr.guest_context->fs = Selector(8, 0, RPL3);
          }

        if ( vm->segmentUpdated & (1<<SRegGS) ) {
          if ( !vm->guest_cpu.desc_cache[SRegGS].valid )
            vm->guest.addr.guest_context->gs = nullSelector.raw;
          else
            vm->guest.addr.guest_context->gs = Selector(9, 0, RPL3);
          }

        for (i=0; i<6; i++) {
          if (vm->guest_cpu.desc_cache[i].valid) {
            vm->guest.addr.gdt[4+i] = vm->guest_cpu.desc_cache[i].desc;
            vm->guest.addr.gdt[4+i].dpl = 3;
            }
          else {
            vm->guest.addr.gdt[4+i] = nullDescriptor;
            }
          }
        vm->segmentUpdated = 0;
        }
      }

    if (context.eflags.fields.vm) {
      /* Copy working sregs back to stack so IRET will use them. */
      v86_sregs.es = context.es;
      v86_sregs.ds = context.ds;
      v86_sregs.fs = context.fs;
      v86_sregs.gs = context.gs;
      vm->guest.addr.tss->esp0 = ((Bit32u)nexus) + PAGESIZE;
      }
    else {
      vm->guest.addr.tss->esp0 =
        ((Bit32u)nexus) + PAGESIZE - sizeof(v86_sregs_t);
      }

    temp8 = 0; /* keep compiler happy */
    CLI();

    /* Set a flag to report back to the assembly code, which mode
     * the guest is being monitored in.  Also, the 16-bit ESP stack
     * hack is setup, when appropriate.  See the notes at the top of
     * this file.
     */
    if (context.eflags.fields.vm) {
      ret = 0; /* to guest in V86M */
      }
    else if ( !vm->guest.addr.gdt[4+SRegSS].d_b ) {
      Bit32u monBaseNormal, monESPUpperNormal, monBaseHacked;
      Bit32u guestESPUpperNormal;
      unsigned ssNormalSlot, ssHackedSlot;
    
      ssNormalSlot = vm->guest.addr.nexus->SSNormal>>3;
      ssHackedSlot = vm->guest.addr.nexus->SS16BitSSHack>>3;
      monBaseNormal = BaseOfDescriptor(vm->guest.addr.gdt[ssNormalSlot]);
    
      asm volatile (
        "movl %%esp, %0 \n\t"
        "andl $0xffff0000, %0"
        : "=r" (monESPUpperNormal)
        );
      guestESPUpperNormal = context.esp & 0xffff0000;
    
      monBaseHacked = monBaseNormal +
                      (monESPUpperNormal - guestESPUpperNormal);
    
      vm->guest.addr.nexus->espUpperNormal = monESPUpperNormal;
      vm->guest.addr.nexus->espUpper16BitSSHack = guestESPUpperNormal;
      vm->guest.addr.gdt[ssHackedSlot] = vm->guest.addr.gdt[ssNormalSlot];
      vm->guest.addr.gdt[ssHackedSlot].base_low  = monBaseHacked;
      vm->guest.addr.gdt[ssHackedSlot].base_med  = monBaseHacked >> 16;
      vm->guest.addr.gdt[ssHackedSlot].base_high = monBaseHacked >> 24;
      ret = 2; /* to guest in PM with SS16 hack */
      }
    else {
      ret = 1; /* to guest in PM */
      }

    if (vm->vOpcodeMap) {
      asm volatile (
        /* Write 'virtual' page table entry, saving the real one. */
        "xchgl %4, (%3)\n\t"

        /* Save byte at beginning of virtualized code page and replace
           it with the opcode for RET, using the alternate address. */
        "movb $0xc3, %5\n\t"
        "xchgb %5, (%1)\n\t"

        /* Invalidate the page entry for the real code page address (I&D). */
        "invlpg (%2)\n\t"

        /* Call the RET instruction (just comes right back) using the actual
           code page address; this loads the TLB (I-only). */
        "call *%2\n\t"

        /* Replace the RET instruction with the original byte, using
           the alternate address. */
        "xchgb %5, (%1)\n\t"

        /* Restore the 'real' page table entry, activating protection. */
        "andl $0xfffffff8, %4\n\t"
        "movl %4, (%3)"

          : "=d" (temp8)
          : "r" (private_codepage_laddr), "r" (vm->codepage_laddr),
            "r" (vm->codepage_pte_p), "r" (pte_new), "0" (temp8)
          : "memory"
        );
      }
    vm->system.t0 = vm_rdtsc();
    return (ret);
}


  unsigned
handle_int( guest_context_t context, v86_sregs_t v86_sregs )
/*
 * handle_int(): Redirect a hardware interrupt back to the host
 */
{
    nexus_t *nexus = (nexus_t *) (((Bit32u) &context) & 0xfffff000);
    vm_t    *vm    = (vm_t *) nexus->vm;
    unsigned from_monitor;

    if (context.eflags.fields.vm) {
      /* Interrupt from guest code executing in v8086 mode. */
      /* The data segment selectors will have been pushed by the */
      /* the CPU for the interrupt.  Copy them down to the area */
      /* where we save them for protected mode. */
      context.es = v86_sregs.es;
      context.ds = v86_sregs.ds;
      context.fs = v86_sregs.fs;
      context.gs = v86_sregs.gs;
      from_monitor = 0; /* Event from guest code */
      }
    else if ( (context.cs & 0x0003) == 0x0003 ) {
      from_monitor = 0; /* Event from guest code */
      }
    else {
      from_monitor = 1; /* Event from monitor code */
      }

    if ( !from_monitor ) {
      Bit64u t1;

      if (vm->vOpcodeMap) { /* SBE enabled */
        /* Reinstate the PTE for the current code page. */
        *vm->codepage_pte_p = vm->codepage_pte_saved;
        /* +++ the following call not really needed. */
        invlpg_mon_offset(vm->codepage_laddr);
        }

      /* End of elapsed guest execution duration.  Add elapsed */
      /* cycles to time framework. */
      t1 = vm_rdtsc();
      vm->system.elapsed += (t1 - vm->system.t0);
      }

    /* Interrupts are off naturally here. */
    vm->mon_request = MON_REQ_REDIRECT;
    vm->redirect_vector = context.vector;
    vm->guest.__mon2host();
    return(from_monitor);
}


  void
handle_fault( guest_context_t context, v86_sregs_t v86_sregs )
/*
 *  handle_fault():  main monitor-side fault handler
 *
 *  This fault handler is called from the assembly stub
 *  __handle_fault.
 *
 *  INPUTS:
 *     context: guest context at point of exception
 *
 */
{
    nexus_t *nexus = (nexus_t *) (((Bit32u) &context) & 0xfffff000);
    vm_t    *vm    = (vm_t *) nexus->vm;
    Bit32u  cr2    = read_cr2();
    Bit64u t1;

    if (context.eflags.fields.vm) {
      /* Interrupt from guest code executing in v8086 mode. */
      /* The data segment selectors will have been pushed by the */
      /* the CPU for the interrupt.  Copy them down to the area */
      /* where we save them for protected mode. */
      context.es = v86_sregs.es;
      context.ds = v86_sregs.ds;
      context.fs = v86_sregs.fs;
      context.gs = v86_sregs.gs;
      }

#if ANAL_CHECKS
    if ( !context.eflags.fields.if_ )
      monpanic(vm, "handle_fault: guest IF=0\n");
    if ( !context.eflags.fields.vm && ((context.cs & 0x0003) != 0x0003) )
      monpanic(vm, "handle_fault: CS.RPL!=3\n");
#endif

    /* End of elapsed guest execution duration */
    t1 = vm_rdtsc();
    vm->system.elapsed += (t1 - vm->system.t0);

    if (vm->vOpcodeMap) { /* SBE enabled */
      /* Reinstate the PTE for the current code page. */
      *vm->codepage_pte_p = vm->codepage_pte_saved;
      invlpg_mon_offset(vm->codepage_laddr);
      }

    STI();

    switch ( context.vector )
    {
    case ExceptionDB: /* 1 */

      if (vm->executeMethod == RunGuestNMethodBreakpoint) {
        /* Breakpoint generated because we requested it via TF=1 */
        if (vm->executeN) {
          vm->executeN--;
          if (vm->executeN==0) {
            sysEOICount(vm);
            }
          }
#warning "breakpoint elapsed hack"
        vm->system.elapsed = 400;
        }
      else {
        monpanic(vm, "handle_fault: #DB, method=%u not coded\n",
          vm->executeMethod);
        }
      break;

    case ExceptionBR: /* 5 */
        /* BOUND instruction fault; array index not in bounds */
        emulate_exception(vm, context.vector, 0);
        break;

    case ExceptionDE: /* 0 */
    case ExceptionBP: /* 3 */
    case ExceptionOF: /* 4 */
    case ExceptionNM: /* 7 */
    case ExceptionMF: /* 16 */
        monpanic(vm, "handle_fault: %u\n", context.vector);
        /* emulate_interrupt(vm, context.vector); */
        break;

    case ExceptionNP: /* 11 */
    case ExceptionSS: /* 12 */
    case ExceptionAC: /* 17 */
        /* use emulate_xyz() */
        /*interrupt(vm, context.vector, 0, 1, context.error); */
        monpanic(vm, "handle_fault: %u\n", context.vector);
        break;

    case ExceptionUD: /* 6 */
    case ExceptionGP: /* 13 */
        emulate_instr(vm, &context, 1);
        break;

    case ExceptionPF: /* 14 */
        monPageFault(vm, &context, cr2);
        break;

    default:
        monpanic(vm, "Other Fault: %u\n", context.vector);
        break;
    }
}

  void
handle_mon_fault( guest_context_t context )
{
  nexus_t *nexus = (nexus_t *) (((Bit32u) &context) & 0xfffff000);
  vm_t    *vm    = (vm_t *) nexus->vm;

  /* Fault occurred inside monitor code; abort. */
  vm->abort_code = 1;
  monpanic_nomess(vm);
}
