/*
 *  plex86: run multiple x86 operating systems concurrently
 *  Copyright (C) 2000  Kevin P. Lawton
 *
 *  mon-mode.c:  Handling of mode to run guest code in.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

/* TODO:
 *   - kick out of RMMonPM when descriptor caches are compat w/ RM
 *     (and switch to differnet virtualization matrix)
 */

#include "plex86.h"
#define IN_MONITOR_SPACE
#include "monitor.h"


/* +++ Deal with desc.valid==0 */
/* +++ Deal with page remapping upon certain transitions */
/* +++ Integrate SBE flag into vm->guestMode */
/* +++ Fix interaction with set_cpu */



unsigned isV86MCompatible(vm_t *);


#warning "need to deal with LGDT for sbe==0"


static unsigned nonSBECompliantCode(vm_t *);


  void
monModeChange(vm_t *vm)
{
  if (vm->modeChange & ModeChangeCheck) {
    /* If the monitor saw an event which may change the mode of
     * the guest, such as a CS register reload, than have a look.
     */
    unsigned newGuestMode;
    if (V8086Mode(vm))
      newGuestMode = GuestModeVM;
    else if (RealMode(vm))
      newGuestMode = GuestModeRM;
    else {
      newGuestMode = GuestModePMR0 | G_GetCPL(vm);
      }
    if (GetGuestMode(vm) != newGuestMode) {
      /* Yep, the guest's mode changed.  Flag this as a transition */
      vm->modeChange |=  ModeChangeTransition;
      }
    vm->modeChange &= ~ModeChangeCheck;
    }

  /* If we don't need remap due to a guest mode change, then
   * skip the following code and at check for the need for a
   * paging remap
   */
  if ( !(vm->modeChange & ModeChangeTransition) )
    goto pageRemapCheck;

  vm->remapHints = 0;

  if (RealMode(vm)) {
    /* Guest ... -> RM */
    switch (GetGuestMode(vm)) {
      case GuestModePMR0:
        /* Guest PMR0 -> RM.  We need to look at
         * the guest descriptor shadow caches.  If they were loaded
         * with RM compatible values, then we can run the RM guest
         * code in v86 mode.  Some software does not load all segment
         * registers with RM compatible values before transitioning
         * back to RM.  As the PM legacy values will be used until
         * a segment reload occurs, this is how some RM code accesses
         * code outside the normal 1Meg+ boundary.
         */

        /* Sanity check: because segment register accesses were previously
         * virtualized, all selectors and shadow cache values should
         * already be maintained in vm->guest_cpu.
         */
        if ( (vm->descriptorInEmu & 0x3f) != 0x3f )
          monpanic(vm, "ModeChange: PMR0->RM, dInEmu=0x%x\n",
            vm->descriptorInEmu);
        if ( (vm->selectorInEmu & 0x3f) != 0x3f )
          monpanic(vm, "ModeChange: PMR0->RM, sInEmu=0x%x\n",
            vm->selectorInEmu);

        if ( isV86MCompatible(vm) ) {
          SetMonGuestMode(vm, MonModeVM, GuestModeRM);
          vm->vOpcodeMap = &vOpcodeMapV86;
          }
        else {
          /* The legacy PM values are not compatible with RM.  Thus,
           * we cannot run this code in VM, because there is no way
           * to maintain the shadow caches values as per the legacy
           * PM values.  Rather we have to use PM, and virtualize
           * segment register accesses.
           */
          SetMonGuestMode(vm, MonModePMR3, GuestModeRM);
          vm->vOpcodeMap = &vOpcodeMapStrongest;
          }
        break;
      default:
        monpanic(vm, "monModeChange mode=%u->RM\n", GetGuestMode(vm));
      }
    }

  else if (ProtectedMode(vm)) {
    /* Guest ... -> PM */
    unsigned currGuestMode;
    currGuestMode = GuestModePMR0 | G_GetCPL(vm);

    if ( G_GetCPL(vm) == 3 ) {
      /* We have PM ring3 code.  If the user has requested SBE turned
       * off for ring3, or has requested Auto-off mode and we have
       * compliant segments, then don't use SBE.
       */

      if ( (vm->prescanRing3 == PrescanRing3Off) ||
           ((vm->prescanRing3 == PrescanRing3Auto ) &&
            nonSBECompliantCode(vm)) ) {
        vm->remapHints |= RemapHintNoSBE;
        }
      }

    /* Since we need to virtualize selectors, move all guest
     * values into guest_cpu area, if not already.
     */
    if ( !(vm->selectorInEmu & (1<<SRegES)) )
      vm->guest_cpu.selector[SRegES].raw =
          vm->guest.addr.guest_context->es;
    if ( !(vm->selectorInEmu & (1<<SRegCS)) )
      vm->guest_cpu.selector[SRegCS].raw =
          vm->guest.addr.guest_context->cs;
    if ( !(vm->selectorInEmu & (1<<SRegSS)) )
      vm->guest_cpu.selector[SRegSS].raw =
          vm->guest.addr.guest_context->ss;
    if ( !(vm->selectorInEmu & (1<<SRegDS)) )
      vm->guest_cpu.selector[SRegDS].raw =
          vm->guest.addr.guest_context->ds;
    if ( !(vm->selectorInEmu & (1<<SRegFS)) )
      vm->guest_cpu.selector[SRegFS].raw =
          vm->guest.addr.guest_context->fs;
    if ( !(vm->selectorInEmu & (1<<SRegGS)) )
      vm->guest_cpu.selector[SRegGS].raw =
          vm->guest.addr.guest_context->gs;
    /* All real selector values moved to guest_cpu area */
    vm->selectorInEmu = 0x3f;

    /* Guest ... -> PM */
    switch (GetGuestMode(vm)) {

      case GuestModeRM:
        /* Guest RM -> PM */
        if (GetMonMode(vm) == MonModePMR3) {
          /* Guest RM(pm) -> PM */
          /* Sanity check: because segment register accesses were previously
           * virtualized, all selectors and shadow cache values should
           * already be maintained in vm->guest_cpu.
           */
          if ( (vm->descriptorInEmu & 0x3f) != 0x3f )
            monpanic(vm, "ModeChange: RM(pm)->PM, dInEmu=0x%x\n",
              vm->descriptorInEmu);
          if ( (vm->selectorInEmu & 0x3f) != 0x3f )
            monpanic(vm, "ModeChange: RM(pm)->PM, sInEmu=0x%x\n",
              vm->selectorInEmu);
          }
        else if (GetMonMode(vm) == MonModeVM) {
          /* Guest RM(vm) -> PM */
          /* Since guest code may have been able to reload segment
           * registers natively, shadow cache values in vm->guest_cpu
           * may not be current.
           */

          /* For now, even in VM, descriptors are virtualized.  So just
           * sanity check here.  But soon, code here should load the
           * vm->guest_cpu shadow registers with values derived from
           * the RM guest selectors
           */
          if ( (vm->descriptorInEmu & 0x3f) != 0x3f )
            monpanic(vm, "ModeChange: RM(vm)->PM, dInEmu=0x%x\n",
              vm->descriptorInEmu);
          if ( (vm->selectorInEmu & 0x3f) != 0x3f )
            monpanic(vm, "ModeChange: RM(vm)->PM, sInEmu=0x%x\n",
              vm->selectorInEmu);
          }
        else
          monpanic(vm, "ModeChange: RM(%u)->PM\n", GetMonMode(vm));
        break;

      case GuestModeVM:
        /* Guest VM -> PM */
        if (GetMonMode(vm) == MonModeVM) {
          /* For now, even in VM, descriptors are virtualized.  So just
           * sanity check here.  But soon, code here should load the
           * vm->guest_cpu shadow registers with values derived from
           * the RM guest selectors
           */
          if ( (vm->descriptorInEmu & 0x3f) != 0x3f )
            monpanic(vm, "ModeChange: VM(vm)->PM, dInEmu=0x%x\n",
              vm->descriptorInEmu);
          if ( (vm->selectorInEmu & 0x3f) != 0x3f )
            monpanic(vm, "ModeChange: VM(vm)->PM, sInEmu=0x%x\n",
              vm->selectorInEmu);
          }
        else
          monpanic(vm, "ModeChange: VM(%u)->PM\n", GetMonMode(vm));
        break;

      case GuestModePMR0:
      case GuestModePMR1:
      case GuestModePMR2:
      case GuestModePMR3:
        /* +++ should have other code here */
        if ( (GetGuestMode(vm) & 0x3) != G_GetCPL(vm) ) {
          /* Previous ring level is not equal to current.  We need
           * to remap paging permissions based on the new level */
          vm->modeChange |= ModeChangePaging;
          }
        else {
          monpanic(vm, "ModeChange: PMR%u->PMR%u\n",
                   GetGuestMode(vm), G_GetCPL(vm));
          }
        break;

      default:
        monpanic(vm, "ModeChange %u->PM\n", GetGuestMode(vm));
      }

    SetMonGuestMode(vm, MonModePMR3, currGuestMode);
    if (vm->remapHints & RemapHintNoSBE)
      vm->vOpcodeMap = 0;
    else
      vm->vOpcodeMap = &vOpcodeMapStrongest;
    }

  else { /* Must be VM */
    /* Guest ... -> VM */
    if ( (GetGuestMode(vm) >= GuestModePMR0) &&
         (GetGuestMode(vm) <= GuestModePMR3) ) {
      /* PMR{0,1,2,3} -> VM */

      /* Sanity check: because segment register accesses were previously
       * virtualized, all selectors and shadow cache values should
       * already be maintained in vm->guest_cpu.
       */
      if ( (vm->descriptorInEmu & 0x3f) != 0x3f )
        monpanic(vm, "ModeChange: PMR0->VM, dInEmu=0x%x\n",
          vm->descriptorInEmu);
      if ( (vm->selectorInEmu & 0x3f) != 0x3f )
        monpanic(vm, "ModeChange: PMR0->VM, sInEmu=0x%x\n",
          vm->selectorInEmu);

      SetMonGuestMode(vm, MonModeVM, GuestModeVM);
      vm->vOpcodeMap = &vOpcodeMapV86;
      }
    else {
      monpanic(vm, "monModeChange mode=%u->VM\n", GetGuestMode(vm));
      }
    }

  vm->remapHints |= RemapHintModeEstablished;

  sysRemapMonitor(vm);

pageRemapCheck:

  /* First set EFLAGS.VM bit according to the mode that the guest
   * will be executed in.
   */
  if (GetMonMode(vm) == MonModeVM)
    vm->guest.addr.guest_context->eflags.fields.vm = 1;
  else
    vm->guest.addr.guest_context->eflags.fields.vm = 0;

  /* If there was a request for a paging remap, handle that now */
  if (vm->modeChange & ModeChangePaging) {
    monPagingRemap(vm);
    }

  /* All mode changes have been handled, reset flag */
  vm->modeChange = 0;
}



/* This function determines if ring3 (user) guest code could be run
 * without SBE intervention.
 */

  unsigned
nonSBECompliantCode(vm_t *vm)
{
  unsigned sreg;

  /* Condition: Guest CPL must equal privilege level that it
   *   is monitored at.  Otherwise, the RPL of segment selectors would
   *   be incorrect and the difference readable by unmonitored guest code.
   *   The calling function should check that the guest is running in
   *   protected mode ring3 before calling.
   */
  VM_ASSERT(vm, G_GetCPL(vm)==3);
  VM_ASSERT(vm, vm->guest_cpu.cr0.fields.pe);
  VM_ASSERT(vm, G_GetVM(vm)==0);

  /* ===================================================================
   * Condition: Guest EFLAGS.IOPL < 3.  This way we can set the monitored
   * IOPL value to the expected guest value, without allowing IOPL
   * sensitive instructions to execute unwantedly.  For example, look at
   * the STI/CLI instructions.  If we let IOPL be 3 (and we know the guest
   * is running at ring3 at this point), then STI/CLI instructions will
   * execute successfully out of control of the VM monitor.  Additionally,
   * IN/OUT instructions would do the same.  As long as we constrain that
   * IOPL < 3, we can set the monitored IOPL value as expected by the
   * guest, thus allowing PUSHF to push expected values.
   */
  if (G_GetIOPL(vm) == 3) {
    InstrNonSBEFail(InstrNonSBEFailIOPL);
    return 0; /* Not compliant */
    }

  /* ===================================================================
   * Condition: Guest EFLAGS.IF must be 1.  If we execute the guest
   * with IF==0, then control may never return to the monitor.  We
   * need to make sure that the next hardware interrupt is received
   * promptly by the monitor and redirected to the host OS.
   */
  if (G_GetIF(vm)==0) {
    InstrNonSBEFail(InstrNonSBEFailIF);
    return 0; /* Not compliant */
    }

  /* ===================================================================
   * Condition: Guest EFLAGS.{VIF,VIP} must be 0 (for now).  I haven't
   * looked into these flags much yet.  So for now, mandate they must
   * be zero.
   */
  if (G_GetVIP(vm) || G_GetVIF(vm)) {
    InstrNonSBEFail(InstrNonSBEFailVIPVIF);
    return 0; /* Not compliant */
    }

  /* ===================================================================
   * Condition: All non-null guest selectors must point to a valid ring3
   * segment descriptor.  Otherwise, the guest selector restore / IRET
   * sequence in the monitor will fail.
   *
   * Condition: The current descriptor cache values must be consistent
   * with the actual descriptor table entries.  Otherwise, the guest
   * selector restore / IRET sequence in the monitor will load values
   * into the segments which are inconsistent with values that the
   * guest has loaded into them.
   */
  if ( (vm->descriptorInEmu & 0x3f) != 0x3f )
    monpanic(vm, "nonSBECC: dInEmu=0x%x\n", vm->descriptorInEmu);
  if ( (vm->selectorInEmu & 0x3f) != 0x3f )
    monpanic(vm, "nonSBECC: sInEmu=0x%x\n", vm->selectorInEmu);

  /* CS and SS selectors must be non-NULL */
  if ( IsNullSelector(vm->guest_cpu.selector[SRegCS]) ) {
    InstrNonSBEFail(InstrNonSBEFailSRegCS);
    return 0; /* Not compliant */
    }
  if ( IsNullSelector(vm->guest_cpu.selector[SRegSS]) ) {
    InstrNonSBEFail(InstrNonSBEFailSRegSS);
    return 0; /* Not compliant */
    }
  for (sreg=0; sreg<6; sreg++) {
    selector_t selector;

    selector = vm->guest_cpu.selector[sreg];
    if ( !IsNullSelector(selector) ) {
      Bit32u descriptor_laddr;
      if ( !vm->guest_cpu.desc_cache[sreg].valid ) {
        InstrNonSBEFail(InstrNonSBEFailSReg + sreg);
        return 0; /* Not compliant */
        }
      /* We have a non-null selector, and it's shadow descriptor cache
       * is marked valid.  Now fetch the corresponding guest descriptor
       * table entry.  This entry must be equivalent to the shadow cache
       * values for our assumptions to work.
       */
      if (selector.fields.ti == 0) { /* GDT */
        if ( (selector.raw | 0x0007) > vm->guest_cpu.gdtr.limit ) {
          InstrNonSBEFail(InstrNonSBEFailGDT);
          return 0; /* Not compliant */
          }
        descriptor_laddr = vm->guest_cpu.gdtr.base +
                           (selector.raw & 0xfff8);
        }
      else { /* LDT */
        if (vm->guest_cpu.ldtr_cache.valid == 0) {
          InstrNonSBEFail(InstrNonSBEFailLDT);
          return 0; /* Not compliant */
          }
        if ( (selector.raw | 0x0007) >
              vm->guest_cpu.ldtr_cache.limit_scaled ) {
          InstrNonSBEFail(InstrNonSBEFailLDT);
          return 0; /* Not compliant */
          }
        descriptor_laddr = vm->guest_cpu.ldtr_cache.base +
                           (selector.raw & 0xfff8);
        }

      if (vm->guest_cpu.cr0.fields.pg) {
        monpanic(vm, "nonSBECC: PG==1 not coded yet.\n");
        }
      else {
        unsigned char *guest_page;
        Bit32u guest_paddr_index;
        descriptor_t *descriptor_p;

        if ( (descriptor_laddr & 0xfff) > 0xffa ) {
          monpanic(vm, "nonSBECC: PG==0 descriptor crosses page.\n");
          }
        guest_paddr_index = A20Addr(vm, descriptor_laddr) >> 12;
        if (guest_paddr_index >= vm->pages.guest_n_pages)
          monpanic(vm, "nonSBECC: paddr_i OOB\n");
        guest_page = open_guest_phy_page(vm, guest_paddr_index,
                                         vm->guest.addr.tmp_phy_page0);
        descriptor_p = (descriptor_t *)
          (guest_page + (descriptor_laddr & 0xfff));

        if ( ((Bit32u *) descriptor_p)[0] !=
             ((Bit32u *) &vm->guest_cpu.desc_cache[sreg].desc)[0] ) {
failSReg:
          monprint(vm, "guest DT %x:%x\n",
            ((Bit32u *) descriptor_p)[0], ((Bit32u *) descriptor_p)[1]);
          monpanic(vm, "cache    %x:%x\n",
            ((Bit32u *) &vm->guest_cpu.desc_cache[sreg].desc)[0],
            ((Bit32u *) &vm->guest_cpu.desc_cache[sreg].desc)[1]);
          InstrNonSBEFail(InstrNonSBEFailSReg + sreg);
          return 0; /* Not compliant */
          }
        if ( ((Bit32u *) descriptor_p)[1] !=
             ((Bit32u *) &vm->guest_cpu.desc_cache[sreg].desc)[1] ) {
          goto failSReg;
          }
        }
      }
    } /* for (...) */

  /* Notes:
   *  - A20 considerations?
   *  - user code access to GDT/LDT/IDT pages?
   */

#if 0
From mon-paging.c:
    /* For guest code running with SBE on, it doesn't matter if */
    /* guest accesses a structure such as GDT, IDT, LDT, since we */
    /* virtualize segment loads and update the A bit immediately. */
    if (!vm->vOpcodeMap) {
      monpanic(vm, "map_guest_laddr: sbe==0\n");
      /* SBE off should only happen when running ring3 guest code, */
      /* which ordinarily doesn't have access to the GDT, IDT, LDT. */
      /* If we add this as a condition before turning off SBE, then */
      /* we don't need to have a check here. */
      }
#endif

  InstrNonSBEPass();
  return 1;
}
